package explicit;

import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Vector;

import parser.State;
import parser.type.TypeBool;
import parser.type.TypeDouble;
import parser.type.TypeInt;
import prism.PrismException;

/**
 * @author Klein
 *
 * A DTMC that arises as the product of DTMCExplicit and
 * some other automaton.
 */

public class DTMCProduct extends DTMCSimple
{
	// the underlying DTMC
	private DTMC dtmc;
	private Object extraInfo;
	private Vector<ProductState> prod_states = new Vector<ProductState>();
	
	public DTMC getDTMC()
	{
		return dtmc;
	}

	public Object getExtraInfo()
	{
		return extraInfo;
	}

	public ProductState getState(int index)
	{
		return prod_states.get(index);
	}
	

	@Override
	public void exportToDotFile(String filename, BitSet mark) throws PrismException
	{
		exportToDotFile(filename, mark, null);
	}
	
	public void exportToDotFile(String filename, BitSet mark, RabinAcceptance acceptance) throws PrismException
	{
		int i;
		try {
			FileWriter out = new FileWriter(filename);
			out.write("digraph " + getModelType() + " {\nsize=\"8,5\"\nnode [shape=box];\n");
			for (i = 0; i < numStates; i++) {
				out.write(i + " [label=\"" + i + " | " + getState(i));
				if (acceptance != null) {
					out.write(" | " + acceptance.getSignature(i));
				}
				// close label
				out.write("\"");
				if (mark != null && mark.get(i))
					out.write(" style=filled  fillcolor=\"#cccccc\"");
				out.write("]\n");
				Iterator<Map.Entry<Integer, Double>> iter = getTransitionsIterator(i);
				while (iter.hasNext()) {
					Map.Entry<Integer, Double> e = iter.next();
					out.write(i + " -> " + e.getKey() + " [ label=\"");
					out.write(e.getValue() + "\" ];\n");
				}
			}
			out.write("}\n");
			out.close();
		} catch (IOException e) {
			throw new PrismException("Could not write " + getModelType() + " to file \"" + filename + "\"" + e);
		}
	}
	
	public ModelCheckerResult projectFromInitialToOriginal(ModelCheckerResult resultProd) {
		ModelCheckerResult result = new ModelCheckerResult();
		result.soln = new double[dtmc.getNumStates()];

		for (Integer state_in_product : getInitialStates()) {
			ProductState p_state = getState(state_in_product);
			Integer state_in_dtmc = p_state.getFirstState();
			
			result.soln[state_in_dtmc] = resultProd.soln[state_in_product];
		}
		
		result.numIters = resultProd.numIters;
		result.timeTaken = resultProd.timeTaken;
		result.timePre = resultProd.timePre;
		result.timeProb0 = resultProd.timeProb0;
		
		return result;
	}
	

	

	static public DTMCProduct generate(final ProductOperator<DTMC> op, boolean allInit) throws PrismException
	{
		final DTMCProduct result = new DTMCProduct();
		final DTMC dtmc = op.getGraph();
		result.dtmc = dtmc;
		result.extraInfo = op.getExtraInfo();
		result.statesList = new ArrayList<State>();

		
		@SuppressWarnings("serial")
		class ProductStateMap extends HashMap<ProductState, Integer>
		{
			public Integer findOrAdd(ProductState state) throws PrismException
			{
				Integer index = get(state);
				if (index == null) {
					index = Integer.valueOf(result.addState());
					put(state, index);
					if (index != result.prod_states.size()) {
						throw new PrismException("Implementation error in product construction!");
					}
					result.prod_states.add(state);
					result.statesList.add(dtmc.getStatesList().get(state.getFirstState()));
					
					// notify operator of the index
					op.notify(state, index);
				}
				return index;
			}
		};
		
		ProductStateMap mapping = new ProductStateMap();
		HashSet<ProductState> expanded = new HashSet<ProductState>();
		LinkedList<ProductState> todo = new LinkedList<ProductState>();
		
		if (allInit) {
			for (int index = 0; index < dtmc.getNumStates(); index++) {
				ProductState prod_initial = op.getInitialState(index);
				Integer s = mapping.findOrAdd(prod_initial);
				todo.add(prod_initial);
				result.addInitialState(s);
			}
		} else {
			for (Integer initial_state : dtmc.getInitialStates()) {
				ProductState prod_initial = op.getInitialState(initial_state);
				Integer s = mapping.findOrAdd(prod_initial);
				todo.add(prod_initial);
				result.addInitialState(s);
			}
		}
		
		while (!todo.isEmpty()) {
			ProductState cur = todo.remove(0);
			if (expanded.contains(cur)) {
				// cur has been expanded in the time between insertion into todo and now
				continue;
			}
			// mark as expanded
			expanded.add(cur);

			Integer from_index = mapping.get(cur);
			
			// expand cur
			Iterator<Entry<Integer, Double>> transitions = dtmc.getTransitionsIterator(cur.getFirstState());
			while (transitions.hasNext()) {
				Entry<Integer, Double> transition = transitions.next();
				
				ProductState to_state = op.getSuccessor(cur, transition.getKey());
				if (!expanded.contains(to_state)) {
					todo.add(to_state);
				}
				Integer to_index = mapping.findOrAdd(to_state);

				result.addToProbability(from_index.intValue(), to_index.intValue(), transition.getValue());
			}
		}
		
		return result;
	}	
	
	StateValues projectStateValuesFromInitialToOriginal(StateValues sv) throws PrismException {
		StateValues result = new StateValues(sv.type, dtmc.getNumStates());
		
		for (Integer initial : getInitialStates()) {
			ProductState product_state = getState(initial);
			
			if (sv.type instanceof TypeBool) {
				result.setBooleanValue(product_state.getFirstState(), (Boolean) sv.getValue(initial));
			} else if (sv.type instanceof TypeInt) {
				result.setIntValue(product_state.getFirstState(), (Integer) sv.getValue(initial));
			} else if (sv.type instanceof TypeDouble) {
				result.setDoubleValue(product_state.getFirstState(), (Double) sv.getValue(initial));
			} else {
				throw new PrismException("Handling for Type "+sv.type+" not implemented.");
			}
		}
		
		return sv;
	}
}
