package explicit;

import java.util.ArrayList;
import java.util.BitSet;

import explicit.rewards.MCRewardsInteger;

import prism.IntegerBound;
import prism.PrismException;

public class DTMCRewardCounterProduct
{
	private DTMCProduct product_graph;
	private ArrayList<BitSet> accRewardToStates;
	private int limit;

	DTMCRewardCounterProduct(int limit) {
		this.limit = limit;
		accRewardToStates = new ArrayList<BitSet>();
		for (int i=0;i<=limit;i++) {
			accRewardToStates.add(new BitSet());
		}
	}
	
	DTMCProduct getDTMC() {
		return product_graph;
	}
	
	/**
	 * Get the states in the product DTMC for a given accumulated reward.
	 * If acc_reward is >= limit, then the states with rewards beyond the
	 * limit are returned.
	 */
	BitSet getStatesWithAccumulatedReward(int acc_reward) {
		if (acc_reward >= limit) {
			return accRewardToStates.get(limit);
		} else {
			return accRewardToStates.get(acc_reward);
		}
	}
	
	/**
	 * Get the states in the product DTMC inside a given integer bound.
	 */
	BitSet getStatesWithAccumulatedRewardInBound(IntegerBound bound) {
		BitSet result = new BitSet();
		for (int r=0; r<=bound.getMaximalInterestingValue(); r++) {
			if (bound.isInBound(r)) {
				result.or(getStatesWithAccumulatedReward(r));
			}
		}
		return result;
	}
	
	/**
	 * Lift the set of states of the original DTMC to the product, i.e.,
	 * from S to  S x Counter for all values of the counter.
	 * @param states
	 * @return
	 */
	BitSet liftStateSet(BitSet states) {
		BitSet result = new BitSet();
		DTMCProduct prod = getDTMC();
		
		for (int i = 0; i < prod.getNumStates(); i++) {
			if (states.get(prod.getState(i).getFirstState())) {
				result.set(i);
			}
		}
		
		return result;
	}
	
	/**
	 * Generate the product of a DTMC with an accumulated reward counter.
	 * The counter has the range [0,limit], with saturation semantics for accumulated
	 * rewards >=limit.
	 * @param graph the DTMC
	 * @param rewards integer MCRewards
	 * @param limit the saturation value for the counter
	 * @param allInit generate product from initial states (false) or from all states (true)
	 * @return
	 * @throws PrismException
	 */
	static public DTMCRewardCounterProduct generate(final DTMC graph, final MCRewardsInteger rewards, final int limit, boolean allInit) throws PrismException {
		final DTMCRewardCounterProduct result = new DTMCRewardCounterProduct(limit);

		class RewardProductOperator implements ProductOperator<DTMC> {
			@Override
			public ProductState getInitialState(Integer graph_state)
			{
				return new ProductState(graph_state, 0);
			}

			@Override
			public ProductState getSuccessor(ProductState from_state, Integer graph_to_state)
			{
				// the accumulated reward
				int acc_reward = from_state.getSecondState();
				
				// reward(s) is accumulated on transition *from* s
				int state_reward = rewards.getStateReward(from_state.getFirstState());
				
				int next_reward = acc_reward + state_reward;
				if (next_reward >= limit) {
					// saturated
					next_reward = limit;
				}
				return new ProductState(graph_to_state, next_reward);
			}

			@Override
			public void notify(ProductState state, Integer index) throws PrismException
			{
				// store product state in the corresponding bitset of accRewardToStates
				int reward = state.getSecondState().intValue();
				result.accRewardToStates.get(reward).set(index.intValue());
			}

			@Override
			public DTMC getGraph()
			{
				return graph;
			}

			@Override
			public Object getExtraInfo()
			{
				return null;
			}
			
		};
		
		RewardProductOperator product_op = new RewardProductOperator();
		
		result.product_graph = DTMCProduct.generate(product_op, allInit);
		
		return result;
	}

}
