package explicit;

import java.util.BitSet;
import java.util.Vector;

import jltl2ba.SimpleLTL;
import parser.ast.Expression;
import parser.ast.ExpressionBinaryOp;
import parser.ast.ExpressionLabel;
import parser.ast.ExpressionTemporal;
import parser.ast.ExpressionUnaryOp;
import parser.type.TypeBool;
import parser.type.TypePathBool;
import prism.Prism;
import prism.PrismException;
import prism.PrismLog;

public class LTLModelChecker
{

	public static LTLProduct<DTMCProduct> getLTLProduct(DTMCModelChecker mc, DTMC dtmc, Expression ltl) throws PrismException
	{
		// Model check maximal state formulas
		Vector<BitSet> labeledStates = new Vector<BitSet>();
		ltl = LTLModelChecker.checkMaximalStateFormulas(mc, dtmc, ltl.deepCopy(), labeledStates);
		
		LTLProduct<DTMCProduct> product = LTLProduct.getProduct(mc.prism, dtmc, ltl, labeledStates, true);
		
		return product;
	}
	
	public static ModelCheckerResult doLTLChecking(DTMCModelChecker mc, DTMC dtmc, Expression ltl) throws PrismException
	{
		LTLProduct<DTMCProduct> product = getLTLProduct(mc, dtmc, ltl);
		
		PrismLog log = mc.getLog();
		// Find accepting maximum end BSCC
		log.println("\nFinding accepting BSCCs...");

		StateValues accepting = mc.findAcceptingBSCCs(product);
		log.println("Number of states in accepting BSCCs = " + accepting.getBitSet().cardinality());

		// Compute reachability probabilities
		log.println("\nComputing reachability probabilities...");
		
		ModelCheckerResult prob_product = mc.computeReachProbs(product.getProductGraph(), accepting.getBitSet());

		// Convert probability vector to original model
		return product.getProductGraph().projectFromInitialToOriginal(prob_product);
	}

	
	/**
	 * Extract maximal state formula from an LTL path formula, model check them (with passed in model checker) and
	 * replace them with ExpressionLabel objects L0, L1, etc. Expression passed in is modified directly, but the result
	 * is also returned. As an optimisation, model checking that results in true/false for all states is converted to an
	 * actual true/false, and duplicate results are given the same label. BitSets giving the states which satisfy each label
	 * are put into the vector  labeledStates, which should be empty when this function is called.
	 */
	public static Expression checkMaximalStateFormulas(StateModelChecker mc, Model model, Expression expr, Vector<BitSet> labeledStates) throws PrismException
	{
		// A state formula
		if (expr.getType() instanceof TypeBool) {
			// Model check
			BitSet sat = mc.checkExpression(model, expr).getBitSet();
			// Detect special cases (true, false) for optimization
			if (sat.isEmpty()) {
				return Expression.False();
			}
			if (sat.equals(model.getAllStateBitSet())) {
				// the sat bitset contains *all* states
				return Expression.True();
			}
			// See if we already have an identical result
			// (in which case, reuse it)
			int i = labeledStates.indexOf(sat);
			if (i != -1) {
				return new ExpressionLabel("L" + i);
			}
			// Otherwise, add result to list, return new label
			int index = labeledStates.size();
			labeledStates.add(sat);
			return new ExpressionLabel("L" + index);
		}
		// A path formula (recurse, modify, return)
		else if (expr.getType() instanceof TypePathBool) {
			if (expr instanceof ExpressionBinaryOp) {
				ExpressionBinaryOp exprBinOp = (ExpressionBinaryOp) expr;
				exprBinOp.setOperand1(checkMaximalStateFormulas(mc, model, exprBinOp.getOperand1(), labeledStates));
				exprBinOp.setOperand2(checkMaximalStateFormulas(mc, model, exprBinOp.getOperand2(), labeledStates));
			} else if (expr instanceof ExpressionUnaryOp) {
				ExpressionUnaryOp exprUnOp = (ExpressionUnaryOp) expr;
				exprUnOp.setOperand(checkMaximalStateFormulas(mc, model, exprUnOp.getOperand(), labeledStates));
			} else if (expr instanceof ExpressionTemporal) {
				ExpressionTemporal exprTemp = (ExpressionTemporal) expr;
				if (exprTemp.getOperand1() != null) {
					exprTemp.setOperand1(checkMaximalStateFormulas(mc, model, exprTemp.getOperand1(), labeledStates));
				}
				if (exprTemp.getOperand2() != null) {
					exprTemp.setOperand2(checkMaximalStateFormulas(mc, model, exprTemp.getOperand2(), labeledStates));
				}
			}
		}
		return expr;
	}

}
