package explicit.conditional;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.BitSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map.Entry;

import parser.State;
import parser.ast.Expression;
import parser.ast.ExpressionConditional;
import parser.ast.ExpressionProb;
import parser.ast.ExpressionTemporal;
import parser.ast.ExpressionUnaryOp;
import prism.PrismException;
import prism.PrismLangException;
import explicit.CTMC;
import explicit.CTMCSimple;
import explicit.DTMC;
import explicit.DTMCSimple;
import explicit.Model;
import explicit.ProbModelChecker;
import explicit.StateValues;

public class DoubleTransformation extends ModelTransformation {

	protected Integer[] stateMappingModeOne;
	protected Integer[] stateMappingModeTwo;
	protected BitSet pivots;

	public DoubleTransformation(final ProbModelChecker mc) {
		super(mc);
	}

	/**
	 * This transformation handles a conditional {@code expression} such that<br/>
	 * 1) the condition is convertible to a simple unbounded until formula and
	 * 2) the expression is a reward expression other than a steady-state reward.
	 * For example: <br/>
	 * <code>
	 * P=? [ F<50 target ][ invariant U target ]<br/>
	 * R=? [ F states][ F target ]
	 * </code>
	 * @param expression
	 * @return true iff instances of this transformation can handle {@code expression}
	 */
	public static boolean instanceCanHandle(final Model model, final ExpressionConditional expression) {
		if (!(model instanceof DTMC)) {
			return false;
		}
		Expression condition = normalizeExpression(expression.getCond());
		if (Expression.isNot(condition)) {
			// currently deactivate handling of negated Until
			return false;
			// can handle negated formula, too
			// condition = ((ExpressionUnaryOp) condition).getOperand();
		}
		try {
			if (!(condition instanceof ExpressionTemporal && condition.isSimplePathFormula())) {
				// can handle simple path conditions only
				return false;
			}
		} catch (PrismLangException e) {
			// expression cannot be checked whether it is a simple formula
			return false;
		}
		final ExpressionTemporal temporal = (ExpressionTemporal) condition;
		if (!(temporal.getOperator() == ExpressionTemporal.P_U)) {
			// can handle until conditions only
			return false;
		}
		if (temporal.hasBounds() || temporal.hasRewardBounds()) {
			// can handle unbounded conditions only
			return false;
		}
		if (new ExpressionInspector(expression.getExpr()).isSteadyStateReward()) {
			// cannot handle steady state computation yet
			return false;
		}

		return true;
	}

	@Override
	public boolean canHandle(final Model model, final ExpressionConditional expression) {
		return instanceCanHandle(model, expression);
	}

	public DTMC transformModel(final DTMC model, final ExpressionConditional expression) throws PrismException {
		super.transformModel(model, expression);

		final Expression condition = normalizeExpression(expression.getCond());
		final double[] probabilities = checkCondition(model, condition);
		final BitSet target = getTargetStates(model, condition);

		return transformModel(model, probabilities, target);
	}

	public DTMC transformModel(DTMC model, double[] probabilities, BitSet target) throws PrismException {
		numberOfStates = model.getNumStates();
		transformedModel = (model instanceof CTMC) ? new CTMCSimple() : new DTMCSimple();
		// FIXME: use ModelExplicit>>copyFrom() ?
		transformStates(model, probabilities, target);
		transformTransitions(model, probabilities);
		transformedModel.setConstantValues(model.getConstantValues());
		transformedModel.findDeadlocks(false);

		// FIXME: check empty state space
		// FIXME: check empty initial state set
		return transformedModel;
	}

	protected double[] checkCondition(final Model model, final Expression condition) throws PrismException {
		final ExpressionProb exprCond = new ExpressionProb(condition, "=", null);

		return mc.checkExpression(model, exprCond).getDoubleArray();
	}

	protected void transformStates(final DTMC model, final double[] probabilities, final BitSet target) {
		// S' = (S_1 x {1}) U (S_2 x {2}) where
		// S_1 is "mode 1" and
		// S_2 is "mode 2"

		final BitSet support = getSupport(probabilities);
		pivots = getPivotStates(model, support, target);
		final BitSet modeOne = getStatesAddedToModeOne(support);
		final BitSet modeTwo = getStatesAddedToModeTwo(model);
		// build state space
		final List<State> stateList = model.getStatesList();
		final List<State> transformedStateList = new ArrayList<State>(modeOne.cardinality() + modeTwo.cardinality());
		stateMappingModeOne = new Integer[numberOfStates];
		stateMappingModeTwo = new Integer[numberOfStates];
		for (int state = 0; state < numberOfStates; state++) {
			if (modeOne.get(state)) {
				// add new state to mode 1
				stateMappingModeOne[state] = transformedModel.addState();
				transformedStateList.add(stateList.get(state));
			}
			if (modeTwo.get(state)) {
				// add new state to mode 2
				stateMappingModeTwo[state] = transformedModel.addState();
				transformedStateList.add(stateList.get(state));
			}
			if (model.isInitialState(state)) {
				addInitialState(state, modeOne, modeTwo);
			}
		}
		transformedModel.setStatesList(transformedStateList);
	}

	protected void addInitialState(int state, BitSet modeOne, BitSet modeTwo) {
		if (modeOne.get(state)) {
			transformedModel.addInitialState(stateMappingModeOne[state]);
		}
		if (modeTwo.get(state) && pivots.get(state)) {
			transformedModel.addInitialState(stateMappingModeTwo[state]);
		}
	}

	protected BitSet getPivotStates(final DTMC model, final BitSet support, final BitSet target) {
		// compute reachable target states == target & support
		BitSet states = (BitSet) target.clone();
		states.and(support);
		return states;
	}

	protected BitSet getStatesAddedToModeOne(final BitSet support) {
		// "mode 1" S_1 = {s from S | Pr(s) > 0} \ target

		// compute states added to mode one == support \ pivots
		final BitSet modeOne = (BitSet) support.clone();
		modeOne.andNot(pivots);
		return modeOne;
	}

	protected BitSet getStatesAddedToModeTwo(final DTMC model) {
		// "mode 2" S_2 = succ*({s from target | Pr(s) > 0})

		// compute states added to mode two == succ*(pivots)
		final BitSet modeTwo = model.computeReachableStates(pivots);
		modeTwo.or(pivots);
		return modeTwo;
	}

	protected BitSet getTargetStates(final DTMC model, final Expression condition) throws PrismException {
		Expression target;
		if (condition instanceof ExpressionUnaryOp) {
			// assume negated until formula
			ExpressionTemporal temporal = (ExpressionTemporal) ((ExpressionUnaryOp) condition).getOperand();
			target = Expression.Not(Expression.Or(temporal.getOperand1(), temporal.getOperand2()));
		} else {
			// assume non-negated until formula
			target = ((ExpressionTemporal) condition).getOperand2();
		}
		return mc.checkExpression(model, target).getBitSet();
	}

	protected void transformTransitions(final DTMC model, final double[] probabilities) {
		for (int state = 0; state < numberOfStates; state++) {
			transformTransitionsModeOne(model, probabilities, state);
			transformTransitionsModeTwo(model, state);
		}
	}

	protected void transformTransitionsModeOne(final DTMC model, final double[] probabilities, int state) {
		if (stateMappingModeOne[state] != null) {
			// Pr(state) > 0 & state is no pivot state
			int mappedState = stateMappingModeOne[state];
			for (Iterator<Entry<Integer, Double>> iter = model.getTransitionsIterator(state); iter.hasNext();) {
				// P'(s,v) = P(s,v) * Pr(v) / Pr(s) for all s in S_1
				Entry<Integer, Double> transition = iter.next();
				int successor = transition.getKey();
				if (probabilities[successor] > 0) {
					int mappedSuccessor = pivots.get(successor) ? stateMappingModeTwo[successor] : stateMappingModeOne[successor];
					double probability = transition.getValue() * probabilities[successor] / probabilities[state];
					transformedModel.setProbability(mappedState, mappedSuccessor, probability);
				}
			}
		}
	}

	protected void transformTransitionsModeTwo(final DTMC model, int state) {
		if (stateMappingModeTwo[state] != null) {
			// state is in succ*(pivots)
			int mappedState = stateMappingModeTwo[state]; 
			for (Iterator<Entry<Integer, Double>> iter = model.getTransitionsIterator(state); iter.hasNext();) {
				// P'(s,v) = P(s,v) for all s in S_2
				Entry<Integer, Double> transition = iter.next();
				int mappedSuccessor = stateMappingModeTwo[transition.getKey()];
				double probability = transition.getValue();
				transformedModel.setProbability(mappedState, mappedSuccessor, probability);
			}
		}
	}

	public StateValues checkExpression(Model model, ExpressionConditional expression) throws PrismException {
		if (!(model instanceof DTMC)) {
			throw new PrismException("Cannot handle " + model.getModelType());
		}
		final DTMC dtmc = (DTMC) model;
		// Debug output
		// final Expression condition = normalizeExpression(expression.getCond());
		// final BitSet target = getTargetStates(dtmc, condition);
		// ModelPrinter.exportToDotFile(dtmc, "conditional/conditional_mc_original.dot", target);
		mc.getLog().println("\nTransforming model (using special handling) for condition: "+expression.getCond());
		long timer = System.currentTimeMillis();
		final DTMC transformedModel = transformModel(dtmc, expression);
		timer = System.currentTimeMillis() - timer;
		mc.getLog().println("Time for model transformation: " + timer / 1000.0 + " seconds.");
		mc.getLog().print("Transformed model has ");
		mc.getLog().println(transformedModel.infoString());
		// Debug output
		// ModelPrinter.exportToDotFile(this, "conditional/conditional_mc_transformed.dot", mapStates(target));
		
		mc.getLog().println("\nComputing property in transformed model ...");
		timer = System.currentTimeMillis();
		final StateValues values = mc.checkExpression(transformedModel, expression.getExpr());
		timer = System.currentTimeMillis() - timer;
		mc.getLog().println("Time for property calculation in transformed model: " + timer / 1000.0 + " seconds.");

		return mapStatesToOrigin(values);
	}

	//----------------- access to transformed model ------------------

	@Override
	public String getStateAttributes(final int i, final BitSet mark) {
		int index = Arrays.asList(stateMappingModeOne).indexOf(i);
		String label = "label=";
		String shape = "shape=";
		String style = "style=\"";
		String fillcolor = "";
	
		if (index >= 0) {
			// mode 1
			label += "\"" + Integer.toString(index) + "\"";
			shape += "box";
		} else {
			// mode 2
			index = Arrays.asList(stateMappingModeTwo).indexOf(i);
			label += "\"" + Integer.toString(index) + "\"";
			shape += "circle";
		}
		if(transformedModel.isInitialState(i)) {
			style += "dashed";
		} else {
			style += "solid";
		}
		if (mark != null && mark.get(i)) {
			style += ",filled";
			fillcolor = "fillcolor=\"#cccccc\"";
		}
		return " [" + label + " " + shape + " " + style + "\" " + fillcolor +"]";
	}

	@Override
	public Integer mapState(final int state) {
		return pivots.get(state) ? stateMappingModeTwo[state]: stateMappingModeOne[state];
	}
}
