package explicit.conditional;

import java.util.BitSet;
import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 * Convenience class to loop easily over the set bits of a BitSet.
 * For example:<br/><br/>
 * <code>for(Integer index : getSetBits(set)) { ... }</code><br/>
 */
public class IterableBitSet implements Iterable<Integer> {

	private BitSet set;

	public IterableBitSet(BitSet set) {
		this.set = set;
	}
	
	@Override
	public Iterator<Integer> iterator() {
		return new Iterator<Integer>() {
			private int index = set.nextSetBit(0);

			@Override
			public boolean hasNext() {
				return index >= 0;
			}

			@Override
			public Integer next() {
				if (hasNext()) {
					Integer next = index;
					index = set.nextSetBit(index + 1);
					return next;
				}
				throw new NoSuchElementException();
			}

			@Override
			public void remove() {
				throw new UnsupportedOperationException();
			}
		};
	}
	
	/**
	 * Convenience method to retrieve the bits of {@code set} that are set.
	 * 
	 * @param set a BitSet
	 * @return an IterableBitSet over the set bits
	 */
	public static IterableBitSet getSetBits(BitSet set) {
		return new IterableBitSet(set);
	}
	
	/**
	 * Simple test method.
	 * 
	 * @param args ignored
	 */
	public static void main(String[] args) {
		BitSet test = new BitSet();
		test.set(1);
		test.set(2);
		test.set(3);
		test.set(5);
		test.set(8);
		test.set(13);
		test.set(21);

		System.out.println("\n" + test + " - set bits:");
		for (Integer index : getSetBits(test)) {
			System.out.println(index);
		}
	}
}
