package explicit.conditional;

import java.util.BitSet;
import java.util.Vector;

import parser.ast.Expression;
import parser.ast.ExpressionConditional;
import prism.LTL2RabinLibrary;
import prism.PrismException;
import explicit.DTMC;
import explicit.DTMCModelChecker;
import explicit.DTMCProduct;
import explicit.DTMCSimple;
import explicit.LTLModelChecker;
import explicit.LTLProduct;
import explicit.Model;
import explicit.ProbModelChecker;
import explicit.StateValues;

public class LTLTransformation extends ModelTransformation {

	protected UntilTransformation transformation;
	protected LTLProduct<DTMCProduct> product;
	protected Integer[] mapping;
	protected BitSet pivots;

	public LTLTransformation(ProbModelChecker mc) {
		super(mc);
	}

	public static boolean instanceCanHandle(final Model model, final ExpressionConditional expression) {
		if (!(model instanceof DTMC)) {
			return false;
		}
		// FIXME: LTL Support check should be possible without expression transformation
		// if(!LTL2RabinLibrary.isSupportedLTL(expression.getCond())) {
		// 	// can handle LTL path conditions only
		// 	return false;
		// }
		// Assume LTLProduct is applicable
		if (new ExpressionInspector(expression.getExpr()).isSteadyStateReward()) {
			// cannot handle steady state computation yet
			return false;
		}

		return true;
	}

	@Override
	public boolean canHandle(final Model model, final ExpressionConditional expression) {
		return instanceCanHandle(model, expression);
	}

	@Override
	public Integer mapState(int state) {
		DTMCProduct model = product.getProductGraph();
		Integer productState = mapping[state];
		return (productState != null) ? transformation.mapState(productState) : null;
	}

	@Override
	public DTMC transformModel(DTMC model, ExpressionConditional expression) throws PrismException {
		super.transformModel(model, expression);
		// FIXME: do not expect mc to be a DTMCModelChecker
		DTMCModelChecker dtmcMc= (DTMCModelChecker) mc;

		product = LTLModelChecker.getLTLProduct(dtmcMc, model, expression.getCond());
		mapping = createMapping(product.getProductGraph());
		pivots = dtmcMc.findAcceptingBSCCs(product).getBitSet();
//		product.getProductGraph().exportToDotFile("conditional/conditional_mc_ltl_product.dot", pivots);
//		ModelPrinter.exportToDotFile(product.getProductGraph(), "conditional/conditional_mc_ltl_product.dot", pivots);

		// FIXME: check type cast
		transformation = new UntilTransformation(mc);
		transformedModel = (DTMCSimple) transformation.transformModel(product.getProductGraph(), pivots);
		return transformedModel;
	}

	protected Integer[] createMapping(DTMCProduct model) {
		Integer[] mapping = new Integer[numberOfStates];
		for (Integer initial : model.getInitialStates()) {
			Integer origin = model.getState(initial).getFirstState();
			assert origin != null : "first state should be set";
			assert mapping[origin] == null : "do not map state twice";
			mapping[origin] = initial;
		}
		return mapping;
	}

	//----------------- access to transformed model ------------------

	@Override
	public String getStateAttributes(final int i, final BitSet mark) {
		int productIndex = transformation.mapStateToOrigin(i);
		int index = product.getProductGraph().getState(productIndex).getFirstState();
		String label = "label=";
		String shape = "shape=";
		String style = "style=\"";
		String fillcolor = "";

		if (! pivots.get(productIndex)) {
			// mode 1
			label += "\"" + Integer.toString(index) + "\"";
			shape += "box";
		} else {
			// mode 2
			label += "\"" + Integer.toString(index) + "\"";
			shape += "circle";
		}
		if(transformedModel.isInitialState(i)) {
			style += "dashed";
		} else {
			style += "solid";
		}
		if (mark != null && mark.get(i)) {
			style += ",filled";
			fillcolor = "fillcolor=\"#cccccc\"";
		}
		return " [" + label + " " + shape + " " + style + "\" " + fillcolor +"]";
	}

	@Override
	public StateValues checkExpression(Model model, ExpressionConditional expression) throws PrismException {
		if (!(model instanceof DTMC)) {
			throw new PrismException("Cannot handle " + model.getModelType());
		}
		final DTMC dtmc = (DTMC) model;
		mc.getLog().println("\nTransforming model (using LTL/DRA) for condition: "+expression.getCond());
		long timer = System.currentTimeMillis();
		final DTMC transformedModel = transformModel(dtmc, expression);
		timer = System.currentTimeMillis() - timer;
		mc.getLog().println("Time for model transformation: " + timer / 1000.0 + " seconds.");
		mc.getLog().print("Transformed model has ");
		mc.getLog().println(transformedModel.infoString());
		// debug output
//		 ModelPrinter.exportToDotFile(dtmc, "conditional/conditional_mc_original.dot", pivots);
//		 ModelPrinter.exportToDotFile(this, "conditional/conditional_mc_transformed.dot", transformation.mapStates(pivots));

		mc.getLog().println("\nComputing property in transformed model ...");
		timer = System.currentTimeMillis();
		final StateValues values = mc.checkExpression(transformedModel, expression.getExpr());
		timer = System.currentTimeMillis() - timer;
		mc.getLog().println("Time for property calculation in transformed model: " + timer / 1000.0 + " seconds.");

		return mapStatesToOrigin(values);
	}
}
