package explicit.conditional;

import java.util.Arrays;
import java.util.BitSet;

import parser.State;
import parser.ast.Expression;
import parser.ast.ExpressionConditional;
import parser.ast.ExpressionProb;
import parser.ast.ExpressionTemporal;
import prism.PrismException;
import prism.PrismLangException;
import explicit.Distribution;
import explicit.MDP;
import explicit.MDPModelChecker;
import explicit.MDPSimple;
import explicit.MDPSparse;
import explicit.Model;
import explicit.ModelCheckerResult;
import explicit.ProbModelChecker;
import explicit.StateValues;

public class MDPTransformation extends ModelTransformation {

	private MDP transformedModel;
	private int fail;
	private int stop;
	private int goal;

	public MDPTransformation(ProbModelChecker mc) {
		super(mc);
	}

	/**
	 * FIXME: adapt comment to MDP reachability Pmax=?(<>E | <>C)
	 * 
	 * This transformation handles a conditional {@code expression} such that<br/>
	 * 1) the condition is convertible to a simple unbounded until formula and
	 * 2) the expression is a reward expression other than a steady-state reward.
	 * For example: <br/>
	 * <code>
	 * P=? [ F<50 target ][ invariant U target ]<br/>
	 * R=? [ F states][ F target ]
	 * </code>
	 * @param expression
	 * @return true iff instances of this transformation can handle {@code expression}
	 */
	public static boolean instanceCanHandle(final Model model, final ExpressionConditional expression) {
		// FIXME: does transformation work with Until as well?
		return instanceCanHandleCondition(expression) && instanceCanHandleExpression(expression);
	}

	private static boolean instanceCanHandleCondition(final ExpressionConditional expression) {
		return isReachabilityExpression(expression.getCond());
	}

	private static boolean instanceCanHandleExpression(final ExpressionConditional expression) {
		if (!(expression.getExpr() instanceof ExpressionProb)) {
			return false;
		}
		String[] relation = {"max=", "<", "<=" };
		if (!Arrays.asList(relation).contains(((ExpressionProb) expression.getExpr()).getRelOp())) {
			// can handle maximal probabilities only
			return false;
		}
		return isReachabilityExpression(((ExpressionProb) expression.getExpr()).getExpression());
	}

	private static boolean isReachabilityExpression(final Expression expression) {
		final Expression normalized = normalizeExpression(expression);
		try {
			if (!(normalized instanceof ExpressionTemporal && normalized.isSimplePathFormula())) {
				// can handle simple path conditions only
				return false;
			}
		} catch (PrismLangException e) {
			// expression cannot be checked whether it is a simple formula
			return false;
		}
		final ExpressionTemporal temporal = (ExpressionTemporal) normalized;
		if (!(temporal.getOperator() == ExpressionTemporal.P_U)) {
			// can handle until conditions only
			return false;
		}
		if (temporal.hasBounds() || temporal.hasRewardBounds()) {
			// can handle unbounded conditions only
			return false;
		}

		return true;
	}

	@Override
	public boolean canHandle(final Model model, final ExpressionConditional expression) {
		return instanceCanHandle(model, expression);
	}

	public MDP transformModel(final MDP model, final ExpressionConditional expression) throws PrismException {
		// super.transformModel(model, expression);
		if(! canHandle(model, expression)) {
			throw new PrismException("Cannot handle " + expression);
		}
		numberOfStates = model.getNumStates();

		// compute E aka "expression goal"
		ExpressionProb expressionProb = (ExpressionProb) expression.getExpr();
		Expression expressionGoal = ((ExpressionTemporal) normalizeExpression(expressionProb.getExpression())).getOperand2();
		BitSet expressionGoalStates = mc.checkExpression(model, expressionGoal).getBitSet();

		// compute C aka "condition goal"
		Expression conditionGoal = ((ExpressionTemporal) normalizeExpression(expression.getCond())).getOperand2();
		BitSet conditionGoalStates = mc.checkExpression(model, conditionGoal).getBitSet();

		// compute B aka "bad states" == {s | Pmin=0(<> (E or C))}
		BitSet targetStates = (BitSet) expressionGoalStates.clone();
		targetStates.or(conditionGoalStates);
		BitSet badStates = ((MDPModelChecker) mc).prob0(model, null, targetStates, true);

		// compute Pmax(<>E | C)
		ModelCheckerResult expressionMaxResult = ((MDPModelChecker) mc).computeReachProbs(model, expressionGoalStates, false);
		double[] expressionMaxProbs = expressionMaxResult.soln;

		// compute Pmax(<>C | E)
		ModelCheckerResult conditionMaxResult = ((MDPModelChecker) mc).computeReachProbs(model, conditionGoalStates, false);
		double[] conditionMaxProbs = conditionMaxResult.soln;

		// copy MDP to new MDPSimple
		MDPSimple transformedModel = null;
		if (model instanceof MDPSimple) {
			transformedModel= new MDPSimple((MDPSimple) model);
		} else if(model instanceof MDPSparse) {
			transformedModel = new MDPSimple((MDPSparse) model);
		} else {
			throw new PrismException("Unsupported MDP class: " + model.getClass());
		}

		// insert states: goal, stop, fail
		// take variables from initial state
		State init = transformedModel.getStatesList().get(transformedModel.getFirstInitialState());
		fail = transformedModel.addState();
		transformedModel.getStatesList().add(init);
		stop = transformedModel.addState();
		transformedModel.getStatesList().add(init);
		goal = transformedModel.addState();
		transformedModel.getStatesList().add(init);

		// redirect choices from expression goal states to goal or fail
		redirectChoices(transformedModel, expressionGoalStates, goal, fail, conditionMaxProbs);
		// redirect choices from condition goal states to goal or stop
		redirectChoices(transformedModel, conditionGoalStates, goal, stop, expressionMaxProbs);

		// add fail choice from bad states to fail
		for (Integer state : new IterableBitSet(badStates)) {
			addDiracChoice(transformedModel, state, fail);
		}

		// add reset choice from fail state to initial state
		addDiracChoice(transformedModel, fail, model.getFirstInitialState());

		// add self-loops to goal, stop
		addDiracChoice(transformedModel, goal, goal);
		addDiracChoice(transformedModel, stop, stop);

		// FIXME: remove states unreachable from init ?
		return (model instanceof MDPSparse) ? new MDPSparse(transformedModel) : transformedModel;
	}

	protected void redirectChoices(MDPSimple model, BitSet states, int target1, int target2, double[] probabilities) {
		for (Integer state : new IterableBitSet(states)) {
			double probability = probabilities[state];
			Distribution redirect = new Distribution();
			if (probability != 0.0) {
				redirect.add(target1, probability);
			}
			if (1.0 - probability != 0.0) {
				redirect.add(target2, 1.0 - probability);
			}
			model.clearState(state);
			model.addChoice(state, redirect);
		}
	}

	protected void addDiracChoice(MDPSimple transformedModel, Integer from, int to) {
		Distribution reset = new Distribution();
		reset.add(to, 1.0);
		transformedModel.addChoice(from, reset);
	}

	protected BitSet getTargetStates(final MDP model, final Expression condition) throws PrismException {
		// assume non-negated until formula
		Expression target = ((ExpressionTemporal) condition).getOperand2();
		return mc.checkExpression(model, target).getBitSet();
	}

	public StateValues checkExpression(Model model, ExpressionConditional expression) throws PrismException {
		if (!(model instanceof MDP)) {
			throw new PrismException("Cannot handle " + model.getModelType());
		}
		final MDP mdp = (MDP) model;
		// debug output
		// final Expression condition = normalizeExpression(expression.getCond());
		// final BitSet target = getTargetStates(mdp, condition);
		// ModelPrinter.exportToDotFile(mdp, "conditional/conditional_mdp_original.dot", target);
		mc.getLog().println("\nTransforming model (using MDP eventually construction) for condition: "+expression.getCond());
		long timer = System.currentTimeMillis();
		transformedModel = transformModel(mdp, expression);
		timer = System.currentTimeMillis() - timer;
		mc.getLog().println("Time for model transformation: " + timer / 1000.0 + " seconds.");
		mc.getLog().print("Transformed model has ");
		mc.getLog().println(transformedModel.infoString());
		// goal state
		final BitSet goal = new BitSet(numberOfStates + 3);
		goal.set(numberOfStates + 2);
		// Debug output
		// ModelPrinter.exportToDotFile(this, transformedModel, "conditional/conditional_mdp_transformed.dot", target);

		// compute probabilities
		mc.getLog().println("\nComputing reachability in transformed model ...");
		timer = System.currentTimeMillis();
		final ModelCheckerResult result = ((MDPModelChecker) mc).computeReachProbs(transformedModel, goal, false);
		timer = System.currentTimeMillis() - timer;
		mc.getLog().println("Time for property calculation in transformed model: " + timer / 1000.0 + " seconds.");

		// FIXME: check bounds
		// FIXME: set undefined values
		return mapStatesToOrigin(StateValues.createFromDoubleArray(result.soln, transformedModel));
	}

	//----------------- access to transformed model ------------------

	@Override
	public String getStateAttributes(int i, BitSet mark) {
		String label = "label=";
		String shape = "shape=";
		String style = "style=\"";
		String fillcolor = "";

		if (i == fail) {
			label += "\"fail\"";
			shape += "circle";
		} else if (i == stop) {
			label += "\"stop\"";
			shape += "circle";
		} else if (i == goal) {
			label += "\"goal\"";
			shape += "circle";
		} else {
			label += "\"" + Integer.toString(i) + "\"";
			shape += "box";
		}
		if(transformedModel.isInitialState(i)) {
			style += "dashed";
		} else {
			style += "solid";
		}
		if (mark != null && mark.get(i)) {
			style += ",filled";
			fillcolor = "fillcolor=\"#cccccc\"";
		}
		return " [" + label + " " + shape + " " + style + "\" " + fillcolor +"]";
	}

	@Override
	public Integer mapState(int state) {
		return state;
	}

}
