package explicit.conditional;

import java.util.BitSet;

import parser.ast.Expression;
import parser.ast.ExpressionConditional;
import parser.ast.ExpressionTemporal;
import parser.ast.ExpressionUnaryOp;
import parser.type.TypeBool;
import parser.type.TypeDouble;
import parser.type.TypeInt;
import prism.PrismException;
import prism.PrismLangException;
import explicit.DTMC;
import explicit.DTMCSimple;
import explicit.Model;
import explicit.ProbModelChecker;
import explicit.StateValues;

abstract public class ModelTransformation {

	protected ProbModelChecker mc;
	protected DTMCSimple transformedModel;
	protected int numberOfStates;

	public ModelTransformation(final ProbModelChecker mc) {
		this.mc = mc;
	}

	/**
	 * Trim surrounding parentheses, double negation and double minus.
	 * If possible, convert the outermost temporal operator to until form.
	 * If it is a negated next operator, move negation to the inner formula.
	 * 
	 * @param expression
	 * @return normalized expression, possibly negated
	 */
	public static Expression normalizeExpression(final Expression expression) {
		Expression normalized = expression.trimUnaryOperations();
		if (normalized instanceof ExpressionTemporal) {
			try {
				return ((ExpressionTemporal) normalized).convertToUntilForm();
			} catch (PrismLangException e) {
				// no until conversion possible
			}
		} else if (Expression.isNot(normalized)) {
			Expression operand = ((ExpressionUnaryOp) normalized).getOperand();
			if (operand instanceof ExpressionTemporal) {
				ExpressionTemporal temporal = (ExpressionTemporal) operand;
				if (temporal.getOperator() == ExpressionTemporal.P_X) {
					// move negation to inner formula
					normalized = Expression.Not(temporal.getOperand2());
					return new ExpressionTemporal(ExpressionTemporal.P_X, null, normalized);
				} else {
					// until conversion, trim double negation if necessary
					normalized = normalizeExpression(temporal);
					return Expression.Not(normalized).trimUnaryOperations();
				}
			}
		}
		return normalized;
	}

	/**
	 * Compute indices i with probabilities(i) > 0.
	 *
	 * @param probabilities
	 * @return support set of probabilities
	 */
	public static BitSet getSupport(final double[] probabilities) {
		// compute indices s with Pr(s) > 0
		final BitSet support = new BitSet(probabilities.length);
		for (int index = 0; index < probabilities.length; index++) {
			if(probabilities[index] > 0) {
				support.set(index);
			}
		}
		return support;
	}

	/**
	 * Delegate method to the class-side test {@code instanceCanHandle()} whether
	 * this transformation type can handle a given conditional expression or not.<br/>
	 * 
	 * FIXME: This method should live on class-side only.
	 * 
	 * @param expression
	 * @return True iff this transformation type can handle the expression.
	 */
	abstract public boolean canHandle(final Model model, final ExpressionConditional expression);

	public DTMC transformModel(final DTMC model, final ExpressionConditional expression) throws PrismException {
		if(! canHandle(model, expression)) {
			throw new PrismException("Cannot handle " + expression);
		}
		numberOfStates = model.getNumStates();
		return null;
	}

	abstract public StateValues checkExpression(Model model, ExpressionConditional expression) throws PrismException;

	//----------------- access to transformed model ------------------

	abstract public String getStateAttributes(final int i, final BitSet mark);

	@Deprecated
	public DTMCSimple getTransformedModel() {
		return transformedModel;
	}

	/**
	 * Map a state index in the original model to the corresponding index in the transformed model.
	 * If there is no corresponding index return {@code null};
	 * 
	 * @param state index in the transformed model
	 * @return index in the original model or {@code null}
	 */
	abstract public Integer mapState(final int state);

	/**
	 * Map a BitSet of states (indices) in the original model to a corresponding set in the transformed model.
	 * Only states that have a corresponding state in the original model are included.
	 * 
	 * @param states indices in the transformed model
	 * @return indices in the original model
	 */
	public BitSet mapStates(final BitSet states) {
		final BitSet mappedStates = new BitSet(numberOfStates);

		for (int state = states.nextSetBit(0); state >= 0 ; state = states.nextSetBit(state + 1)) {
			Integer mappedState = mapState(state);
			if (mappedState != null) {
				mappedStates.set(mappedState);
			}
		}
		return mappedStates;
	}

	/**
	 * Map StateValues for states (indices) in the transformed model to StateValues for the corresponding states in the original model.
	 * Only values for states that have a corresponding state in the original model are set.
	 * 
	 * @param values indices in the transformed model
	 * @return
	 */
	public StateValues mapStatesToOrigin(final StateValues values) throws PrismLangException {
		if (values.getBitSet() != null) {
			return mapStatesToOrigin(TypeBool.getInstance(), values);
		}
		if (values.getIntArray() != null) {
			return mapStatesToOrigin(TypeInt.getInstance(), values);
		}
		if (values.getDoubleArray() != null) {
			return mapStatesToOrigin(TypeDouble.getInstance(), values);
		}
		assert false : "unknow type";
		return null;
	}

	protected StateValues mapStatesToOrigin(final TypeBool type, final StateValues values) throws PrismLangException {
		final StateValues mappedValues = new StateValues(type, numberOfStates);
		final BitSet data = values.getBitSet();

		for (int state = 0; state < numberOfStates; state++) {
			Integer mappedState = mapState(state);
			if (mappedState == null) {
				// FIXME: choose sane default value
				mappedValues.setBooleanValue(state, false);
			} else {
				mappedValues.setBooleanValue(state, data.get(mappedState));
			}
		}
		return mappedValues;
	}

	protected StateValues mapStatesToOrigin(final TypeInt type, final StateValues values) throws PrismLangException {
		final StateValues mappedValues = new StateValues(type, numberOfStates);
		final int[] data = values.getIntArray();

		for (int state = 0; state < numberOfStates; state++) {
			Integer mappedState = mapState(state);
			if (mappedState == null) {
				// FIXME: choose sane default value
				mappedValues.setIntValue(state, -1);
			} else {
				mappedValues.setIntValue(state, data[mappedState]);
			}
		}
		return mappedValues;
	}

	protected StateValues mapStatesToOrigin(final TypeDouble type, final StateValues values) throws PrismLangException {
		final StateValues mappedValues = new StateValues(type, numberOfStates);
		final double[] data = values.getDoubleArray();

		for (int state = 0; state < numberOfStates; state++) {
			Integer mappedState = mapState(state);
			if (mappedState == null) {
				// FIXME: choose sane default value
				mappedValues.setDoubleValue(state, -1);
			} else {
				mappedValues.setDoubleValue(state, data[mappedState]);
			}
		}
		return mappedValues;
	}
}
