package explicit.conditional;

import java.util.BitSet;
import java.util.Iterator;
import java.util.Map.Entry;

import parser.ast.Expression;
import parser.ast.ExpressionConditional;
import parser.ast.ExpressionTemporal;
import prism.PrismException;
import prism.PrismLangException;
import explicit.DTMC;
import explicit.Model;
import explicit.ProbModelChecker;

public class NextTransformation extends DoubleTransformation {

	public NextTransformation(final ProbModelChecker mc) {
		super(mc);
	}

	/**
	 * This transformation handles a conditional {@code expression} such that<br/>
	 * 1) the condition is (possibly negated) next formula and
	 * 2) the expression is a reward expression other than a steady-state reward.
	 * For example: <br/>
	 * <code>
	 * P=? [ F<50 states ][ X target ]<br/>
	 * R=? [ F states][ ! X target ]
	 * </code>
	 * @param expression
	 * @return true iff instances of this transformation can handle {@code expression}
	 */
	public static boolean instanceCanHandle(final Model model, final ExpressionConditional expression) {
		if (!(model instanceof DTMC)) {
			return false;
		}
		final Expression condition = normalizeExpression(expression.getCond());
		try {
			if (!(condition instanceof ExpressionTemporal && condition.isSimplePathFormula())) {
				// can handle simple path conditions only
				return false;
			}
		} catch (PrismLangException e) {
			// expression cannot be checked whether it is a simple formula
			return false;
		}
		final ExpressionTemporal temporal = (ExpressionTemporal) condition;
		if (!(temporal.getOperator() == ExpressionTemporal.P_X)) {
			// can handle next conditions only
			return false;
		}
		if (temporal.hasBounds() || temporal.hasRewardBounds()) {
			// can handle unbounded conditions only
			return false;
		}
		if (new ExpressionInspector(expression.getExpr()).isSteadyStateReward()) {
			// cannot handle steady state computation yet
			return false;
		}

		return true;
	}

	@Override
	public boolean canHandle(final Model model, final ExpressionConditional expression) {
		return instanceCanHandle(model, expression);
	}

	@Override
	protected void addInitialState(int state, BitSet modeOne, BitSet modeTwo) {
		if (modeOne.get(state)) {
			transformedModel.addInitialState(stateMappingModeOne[state]);
		}
	}

	@Override
	protected BitSet getPivotStates(final DTMC model, final BitSet support, final BitSet target) {
		// compute reachable target states == target & succ(support)
		BitSet states = model.computeSuccessors(support);
		states.and(target);
		return states;
	}

	@Override
	protected BitSet getStatesAddedToModeOne(final BitSet support) {
		// "mode 1" S_1 = {s from S | Pr(s) > 0}

		// compute states added to mode one == support
		final BitSet modeOne = (BitSet) support.clone();
		return modeOne;
	}

	@Override
	protected BitSet getTargetStates(final DTMC model, final Expression condition) throws PrismException {
		// assume condition is an ExpressionTemporal
		return mc.checkExpression(model, ((ExpressionTemporal) condition).getOperand2()).getBitSet();
	}

	@Override
	protected void transformTransitionsModeOne(final DTMC model, final double[] probabilities, int state) {
		if (stateMappingModeOne[state] != null) {
			// Pr(state) > 0 & state is no pivot state
			int mappedState = stateMappingModeOne[state];
			for (Iterator<Entry<Integer, Double>> iter = model.getTransitionsIterator(state); iter.hasNext();) {
				// P'(s,v) = P(s,v) / Pr(s) for all s in S', v in target
				Entry<Integer, Double> transition = iter.next();
				int successor = transition.getKey();
				if (pivots.get(successor)) {
					int mappedSuccessor = stateMappingModeTwo[successor];
					double probability = transition.getValue() / probabilities[state];
					transformedModel.setProbability(mappedState, mappedSuccessor, probability);
				}
			}
		}
	}

	//----------------- access to transformed model ------------------

	@Override
	public Integer mapState(final int state) {
		return stateMappingModeOne[state];
	}
}
