package explicit.conditional;

import java.util.BitSet;

import parser.ast.Expression;
import parser.ast.ExpressionConditional;
import parser.ast.ExpressionReward;
import parser.ast.ExpressionTemporal;
import explicit.DTMC;
import explicit.Model;
import explicit.ProbModelChecker;

public class TerminalTransformation extends DoubleTransformation {

	public TerminalTransformation(final ProbModelChecker mc) {
		super(mc);
	}

	/**
	 * This transformation handles a conditional reachability reward {@code expression} such that<br/>
	 * 1) the condition is convertible to a simple unbounded until and
	 * 2) the reward goal equals the until goal.
	 * For example: <br/>
	 * <code>
	 * R=? [ F target ][ invariant U target]
	 * </code>
	 * @param expression
	 * @return true iff instances of this transformation can handle {@code expression}
	 */
	public static boolean instanceCanHandle(final Model model, final ExpressionConditional expression) {
		if (! DoubleTransformation.instanceCanHandle(model, expression)) {
			return false;
		}

		final Expression condition = normalizeExpression(expression.getCond());
		if (!(condition instanceof ExpressionTemporal)) {
			// can handle non-negated conditions only
			return false;
		}
		if (! new ExpressionInspector(expression.getExpr()).isReachablilityReward()) {
			// can handle reachability rewards only
			return false;
		}
		
		final ExpressionReward reward = (ExpressionReward) expression.getExpr();
		final Expression rewardGoal = ((ExpressionTemporal) reward.getExpression()).getOperand2();
		final Expression conditionGoal = ((ExpressionTemporal) condition).getOperand2();
		return rewardGoal.equals(conditionGoal);
	}
	
	@Override
	public boolean canHandle(final Model model, final ExpressionConditional expression) {
		return instanceCanHandle(model, expression);
	}

	@Override
	protected BitSet getStatesAddedToModeTwo(DTMC model) {
		// "mode 2" S_2 = {s from target | Pr(s) > 0}

		// compute states added to mode two == pivots
		final BitSet modeTwo = pivots;
		return modeTwo;
	}

	@Override
	protected void transformTransitionsModeTwo(final DTMC model, int state) {
		// transitions in mode 2
		if (stateMappingModeTwo[state] != null) {
			// state is in succ*(pivots) == pivots
			// P'(s,s) = 1 for all s in pivot set
			int mappedState = stateMappingModeTwo[state]; 
			transformedModel.setProbability(mappedState, mappedState, 1);
		}
	}
}
